%% CS294A/CS294W Convolutional Neural Networks Exercise

%  Instructions
%  ------------
% 
%  This file contains code that helps you get started on the
%  convolutional neural networks exercise. 

%%======================================================================
%% STEP 0: Initialization
%  Here we initialize some parameters used for the exercise.

imageDim = 64;         % image dimension
imageChannels = 3;     % number of channels (rgb, so 3)

patchDim = 8;          % patch dimension
numPatches = 100000;   % number of patches

visibleSize = patchDim * patchDim * imageChannels;  % number of input units 
outputSize = visibleSize;   % number of output units
hiddenSize = 400;           % number of hidden units 

outputDir = '.';
sparsityParam = 0.035; % desired average activation of the hidden units.
lambda = 3e-3;         % weight decay parameter       
beta = 5;             % weight of sparsity penalty term       

epsilon = 0.1;	       % epsilon for ZCA whitening

poolDim = 22;          % dimension of pooling region

global AEOptions;
AEOptions.linearAutoencoder = true;

%%======================================================================
%% STEP 1: Train a sparse autoencoder (with a linear decoder) to learn 
%  features from color patches. If you have completed the linear decoder
%  execise, use the features that you have obtained from that exercise, 
%  loading them into optTheta. Recall that we have to keep around the 
%  parameters used in whitening (i.e., the ZCA whitening matrix and the
%  meanPatch)

% --------------------------- YOUR CODE HERE --------------------------
% Train the sparse autoencoder and fill the following variables with 
% the optimal parameters:

optTheta =  zeros(2*hiddenSize*visibleSize+hiddenSize+visibleSize, 1);
ZCAWhite =  zeros(visibleSize, visibleSize);
meanPatch = zeros(visibleSize, 1);

load stlSampledPatches.mat 

displayColorNetwork(patches(:, 1:100));

%% STEP 2b: Apply preprocessing
%  In this sub-step, we preprocess the sampled patches, in particular, 
%  ZCA whitening them. Note that because you will have to duplicate the
%  preprocessing steps EXACTLY later, you should store the meanPatch and
%  ZCAWhite matrices used to preprocess the patches you learn the features
%  on somewhere.

% Subtract mean patch (hence zeroing the mean of the patches)
meanPatch = mean(patches, 2);  
patches = bsxfun(@minus, patches, meanPatch);

% Apply ZCA whitening
sigma = patches * patches' / numPatches;
[u, s, v] = svd(sigma);
ZCAWhite = u * diag(1 ./ (diag(s) + epsilon)) * u';
patches = ZCAWhite * patches;

displayColorNetwork(patches(:, 1:100));

%% STEP 2c: Learn features
%  You will now use your sparse autoencoder (with linear decoder) to learn
%  features on the preprocessed patches. This should take around 30 minutes.

theta = initializeParameters(hiddenSize, visibleSize);

% Use minFunc to minimize the function
addpath minFunc/

clear('options');
options.Method = 'lbfgs'; 
options.maxIter = 400;
options.display = 'on';

startTime = tic();
[optTheta, cost] = minFunc( @(p) sparseAutoencoderLinearCost(p, ...
                                   visibleSize, hiddenSize, ...
                                   lambda, sparsityParam, ...
                                   beta, patches), ...
                              theta, options);
duration = toc(startTime);
fprintf('Learning features took %0.3f seconds\n', duration);

% --------------------------------------------------------------------

% Display and check to see that the features look good
W = reshape(optTheta(1:visibleSize * hiddenSize), hiddenSize, visibleSize);
b = optTheta(2*hiddenSize*visibleSize+1:2*hiddenSize*visibleSize+hiddenSize);

displayColorNetwork( (W*ZCAWhite)');


%%======================================================================
%% STEP 3: Convolution and pooling
%  In this step, you will convolve each of the features you learned with
%  the full large images to obtained the convolved features. You will then
%  pool the convolved features to obtain the pooled features for
%  classification.
%
%  Tip: when first writing and testing your convolution, you might want to
%  reduce the number of images you convolve with. This can be accomplished
%  by passing an additional parameter to loadSTLData, as described in step 2a.


%% STEP 3a: Convolution
%  Implement convolution in the function cnnConvolve in cnnConvolve.m

% We have to preprocess the images in the exact same way we preprocessed
% the patches before we can obtain the feature activations.

for convPart = 1:8
    
fprintf('*** convPart = %d\n', convPart);
    
Wt = W( (convPart-1) * 50 + 1:convPart * 50, :);
bt = b( (convPart-1) * 50 + 1:convPart * 50);
hiddenSize = 50;

[numImages, images, labels] = loadSTL10Data('trainSmall64');
images = images / 255;

% NOTE: Implement cnnConvolve in cnnConvolve.m first!
startTime = tic();
convolvedFeatures = cnnConvolve(patchDim, hiddenSize, images, Wt, bt, ZCAWhite, meanPatch);
duration = toc(startTime);

fprintf('Convolution took %0.3f seconds\n', duration);
save('results/convolvedFeaturesC.mat', 'convolvedFeatures', 'duration');

%{
%% STEP 3b: Checking your convolution
%  To ensure that you have convolved the features correctly, we have
%  provided some code to compare the results of your convolution with
%  activations from the sparse autoencoder

% For 1000 random points
for i = 1:1000    
    featureNum = randi([1, hiddenSize]);
    imageNum = randi([1, numImages]);
    imageRow = randi([1, imageDim - patchDim + 1]);
    imageCol = randi([1, imageDim - patchDim + 1]);    
   
    patch = images(imageRow:imageRow + patchDim - 1, imageCol:imageCol + patchDim - 1, :, imageNum);
    patch = patch(:);            
    patch = patch - meanPatch;
    patch = ZCAWhite * patch;
    
    [~, ~, features] = sparseAutoencoderCost(sparseAEModel.optTheta, visibleSize, hiddenSize, ...
                                             lambda, sparsityParam, beta, patch);

    if abs(features(featureNum, 1) - convolvedFeatures(featureNum, imageNum, imageRow, imageCol)) > 1e-9
        fprintf('Convolved feature does not match activation from autoencoder\n');
        fprintf('Feature Number    : %d\n', featureNum);
        fprintf('Image Number      : %d\n', imageNum);
        fprintf('Image Row         : %d\n', imageRow);
        fprintf('Image Column      : %d\n', imageCol);
        fprintf('Convolved feature : %0.5f\n', convolvedFeatures(featureNum, imageNum, imageRow, imageCol));
        fprintf('Sparse AE feature : %0.5f\n', features(featureNum, 1));       
        error('Convolved feature does not match activation from autoencoder');
    end 
end

disp('Check passed');
%}
disp('Did not check.');

%% STEP 3c: Pooling
%  Implement pooling in the function cnnPool in cnnPool.m

for poolDim = [6 16]

    fprintf('poolDim = %d\n', poolDim);
    % NOTE: Implement cnnPool in cnnPool.m first!
    pooledFeatures = cnnPool(poolDim, convolvedFeatures);
    save(sprintf('results/pooledFeaturesC-%d-%d.mat', poolDim, convPart), 'pooledFeatures', 'poolDim', 'convPart');
    
end

%% Convolution and pooling for test set

[numImages, images, testLabels] = loadSTL10Data('testSmall64');
images = images / 255;

startTime = tic();
convolvedFeatures = cnnConvolve(patchDim, hiddenSize, images, Wt, bt, ZCAWhite, meanPatch);
duration = toc(startTime);
fprintf('Convolution took %0.3f seconds\n', duration);
save('results/convolvedFeaturesTestC.mat', 'convolvedFeatures', 'duration');

%%

for poolDim = [6 16]

    fprintf('poolDim = %d\n', poolDim);
    % NOTE: Implement cnnPool in cnnPool.m first!
    pooledFeatures = cnnPool(poolDim, convolvedFeatures);
    save(sprintf('results/pooledFeaturesTestC-%d-%d.mat', poolDim, convPart), 'pooledFeatures', 'poolDim', 'convPart');
    
end

disp('Moving on...');

end

error('Done for now');

%% Consolidate pooled feature parts

for poolDim = [6 16]
    fprintf('poolDim = %d\n', poolDim);
    
    hiddenSize = 50;
    numImages = 3200;
    pooledFeatures = zeros(400, numImages, floor(57 / poolDim), floor(57 / poolDim));
    
    for convPart = 1:2
        fprintf('\tconvPart = %d\n', convPart);
        pooledFeaturesPart = load(sprintf('results/pooledFeaturesTestC-%d-%d.mat', poolDim, convPart));
        pooledFeaturesPart = pooledFeaturesPart.pooledFeatures;
        pooledFeatures((convPart-1)*50+1:(convPart)*50, :, :, :) = pooledFeaturesPart;
    end    
    
    fprintf('Saving %d\n', poolDim);
    save(sprintf('results/pooledFeaturesTestC-%d-all.mat', poolDim), 'pooledFeatures', 'poolDim');
end

%%======================================================================
%% STEP 4: Use pooled features for classification
%  Now, you will use your pooled features to train a softmax classifier,
%  using softmaxTrain from the softmax exercise.
%  Training the softmax classifer for 1000 iterations should take less than
%  10 minutes.

% Add the path to your softmax solution, if necessary
% addpath /path/to/solution/

addpath ../../softmax/soln

for pooledDim = [6 16]
%for pooledDim = [6 8 12 16 32]
    
% for sml = [1e-4 5e-4 1e-3 5e-3 1e-2]
for sml = [ 1e-3 1e-2]

pooledFeatures = load(sprintf('results/pooledFeaturesC-%d-all.mat', pooledDim), 'pooledFeatures');
pooledFeatures = pooledFeatures.pooledFeatures;
numImages = 2000;
    
fprintf('pooledDim = %d, softmax lambda: %g\n', pooledDim, sml);
    
% Setup parameters for softmax
% softmaxLambda = 1e-6;
softmaxLambda = sml;

numClasses = 4;
% Reshape the pooledFeatures to form an input vector for softmax
softmaxX = permute(pooledFeatures, [1 3 4 2]);
softmaxX = reshape(softmaxX, numel(pooledFeatures) / numImages, numImages);
softmaxY = labels;

startTime = tic();
clear('options');
options.display = 'final';
options.maxIter = 500;
softmaxModel = softmaxTrain(numel(pooledFeatures) / numImages, numClasses, softmaxLambda, softmaxX, softmaxY, options);
duration = toc(startTime);

fprintf('Training softmax took %0.3f seconds\n', duration);
save(sprintf('results/softmaxModelC-%d-%g.mat', pooledDim, sml), 'softmaxModel', 'duration', 'softmaxLambda', 'pooledDim');

%%======================================================================
%% STEP 5: Test classifer
%  Now you will test your trained classifer on the test images

pooledFeatures = load(sprintf('results/pooledFeaturesTestC-%d-all.mat', pooledDim),'pooledFeatures');
pooledFeatures = pooledFeatures.pooledFeatures;
numImages = 3200;

softmaxX = permute(pooledFeatures, [1 3 4 2]);
softmaxX = reshape(softmaxX, numel(pooledFeatures) / numImages, numImages);
softmaxY = testLabels;

[pred, prob] = softmaxPredict(softmaxModel, softmaxX);
acc = (pred(:) == softmaxY(:));
acc = sum(acc) / size(acc, 1);
fprintf('Accuracy: %2.3f%%\n', acc * 100);

% You should expect to get an accuracy of around 50% on the test images.

end

end
