function [] = displayPatches(patches, patchRows, patchCols, patchChannels)
%displayPatches(patches, patchRows, patchCols, patchChannels)
% patches - patchRows*patchCols*patchChannels x numPatches matrix
% patchRows - size of patch - number of rows
% patchCols - size of patch - number of columns
% patchChannels - number of (colour) channels

numPatches = size(patches, 2);
dispRows = floor(sqrt(numPatches));
dispCols = dispRows;
dispPatches = dispRows * dispCols;

patches = reshape(patches, patchRows, patchCols, patchChannels, numPatches);
borderWidth = 1;

canvas = zeros(dispRows * patchRows + (dispRows - 1) * borderWidth, ...
               dispCols * patchCols + (dispCols - 1) * borderWidth, ...
               patchChannels);

% Fill in patches
for dRow = 1:dispRows
    for dCol = 1:dispCols
        patchNum = (dRow - 1) * dispCols + dCol;
        
        cRow = (dRow - 1) * (patchRows + borderWidth);
        cCol = (dCol - 1) * (patchCols + borderWidth);
        
        canvas(cRow + 1:cRow + patchRows, cCol + 1:cCol + patchCols, :) = patches(:,:,:,patchNum);
        
        % Fill in border 
        if dRow > 1
            canvas(cRow + 1 - borderWidth:cRow, :, :) = max(abs(patches(:)));
        end
        if dCol > 1
            canvas(:, cCol + 1 - borderWidth:cCol, :) = max(abs(patches(:)));
        end
        
    end
end

canvas = canvas ./ max(abs(canvas(:)));
canvas = canvas ./ 2;
canvas = canvas + 0.5;

handle = image(canvas);
axis off;

end

