function patches = sampleImages(IMAGES, patchSize, numChannels, numPatches)

% sampleImages
% Returns numPatches patches for training

% Initialize patches with zeros.  Your code will fill in this matrix--one
% column per patch, numPatches columns. 
patches = zeros(patchSize*patchSize*numChannels, numPatches);

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Fill in the variable called "patches" using data 
%  from IMAGES.  
%  
%  IMAGES(r, c, channel, which image)
%  so IMAGES(:, :, :, 2) is the 2nd image
%  and IMAGES(r, c, 2, 4) is the 2nd channel's colour at (r, c) for the 4th image

% ------------ Sample Solution ------------
p = 0;
numImages = size(IMAGES, 4);
for im = 1:numImages
    
    % Sample Patches
    numSamples = numPatches / numImages;
    
    for s = 1:numSamples
        acceptable = false;
        
        % DEBUG (for photos): 
        % Reject low-variance images (to avoid entirely white patches); 
        % Reject high mean images (to avoid patches of sky);
        p = p + 1;        
        
        varmin = 500;
        while ~acceptable            
            y = randi(size(IMAGES,1)-patchSize+1);
            x = randi(size(IMAGES,2)-patchSize+1);
            sample = double(IMAGES(y:y+patchSize-1, x:x+patchSize-1, :, im));
            patches(:, p) = sample(:);
            if ((var(sample(:)) < varmin) || (mean(sample(:)) > 256))
                acceptable = false;
                varmin = varmin * 0.99;
            else                
                acceptable = true;
            end
        end
        
    end
    
    fprintf('%d\n', im);
   
end
% ---------- End Sample Solution ---------- 

end


%% ---------------------------------------------------------------
function patches = normalizeData(patches)

% Squash data to [0.1, 0.9] since we use sigmoid as the activation
% function in the output layer

% Remove DC (mean of images). 
patches = bsxfun(@minus, patches, mean(patches));

% Truncate to +/-3 standard deviations and scale to -1 to 1
pstd = 3 * std(patches(:));
patches = max(min(patches, pstd), -pstd) / pstd;

% Rescale from [-1,1] to [0.1,0.9]
patches = (patches + 1) * 0.4 + 0.1;

end
