function [cost, grad] = softmaxCost(theta, numClasses, inputSize, lambda, data, labels)

% numClasses - the number of classes 
% inputSize - the size N of the input vector
% lambda - weight decay parameter
% data - the N x M input matrix, where each column data(:, i) corresponds to
%        a single test set
% labels - an M x 1 matrix containing the labels corresponding for the input data
%

% Unroll the parameters from theta
theta = reshape(theta, numClasses, inputSize);

M = size(data, 2);
groundTruth = full(sparse(labels, 1:M, 1));
cost = 0;

thetagrad = zeros(numClasses, inputSize);

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Compute the cost and gradient for softmax regression.
%                You need to compute thetagrad and cost.
%                The groundTruth matrix might come in handy.

thetaMatrix = theta * data;
thetaMatrix = bsxfun(@minus, thetaMatrix, max(thetaMatrix, [],  1)); % rescale
expThetaMatrix = exp(thetaMatrix);
sumMatrix = sum(expThetaMatrix, 1);

predMatrix = bsxfun(@rdivide, expThetaMatrix, sumMatrix);

loss = (thetaMatrix .* groundTruth);
loss = (sum(loss(:)) - sum(log(sumMatrix(:)), 1)) / size(data, 2);

cost = loss;
thetagrad = (groundTruth - predMatrix) * data' / size(data, 2);

cost = -cost + lambda / 2 * sum(theta(:).^2);
thetagrad = -thetagrad + lambda * theta;

% Unroll the gradient matrices into a vector for minFunc
grad = [thetagrad(:)];
end

