function [cost, grad] = sparseCodingCost(which, theta, visibleSize, numFeatures, patches, gamma, lambda, epsilon)
%sparseCodingCost 
%   which - 'weight', 'feature' or 'both'

    numExamples = size(patches, 2);

    weightMatrix = reshape(theta(1:visibleSize * numFeatures), visibleSize, numFeatures);
    featureMatrix = reshape(theta(visibleSize * numFeatures + 1:visibleSize * numFeatures + numFeatures * numExamples), numFeatures, numExamples);

    % Reconstruction error
    cost = weightMatrix * featureMatrix - patches;
    cost = sum(cost(:) .^ 2); 
    
    % Be careful! This is a differentiable form of L1-penalty 
    % (__NOT__ an L2 penalty), hence the sqrt
    sparsityPenaltyMatrix = sqrt(featureMatrix .^ 2 + epsilon);
    sparsityPenalty = sum(sparsityPenaltyMatrix(:));
    
    % Basis weight penalty
    cost = cost + (gamma / 2) * sum(weightMatrix(:) .^ 2);
    
    % Feature sparsity penalty
    cost = cost + lambda * sparsityPenalty;    
    
    delta = weightMatrix * featureMatrix - patches;
    weightGrad = 2 * delta * featureMatrix'; 
    weightGrad = weightGrad + gamma * weightMatrix;
    featureGrad = 2 * weightMatrix' * delta;    
    featureGrad = featureGrad + lambda * featureMatrix ./ sparsityPenaltyMatrix;
    
    if isequal(which, 'weight')
        featureGrad = featureGrad * 0;
    elseif isequal(which, 'feature')
        weightGrad = weightGrad * 0;
        disp(cost);
        disp(lambda * sparsityPenalty);
    end
    
    grad = [weightGrad(:); featureGrad(:)];
    grad = grad(:);        
end