%% CS294A/CS294W Sparse Coding Exercise

%  Instructions
%  ------------
% 
%  This file contains code that helps you get started on the
%  sparse coding exercise. In this exercise, you will need to modify
%  FILENAME.m.

addpath ../../vectorization/soln
addpath ../../minFunc
addpath ../../minConf

%%======================================================================
%% STEP 0: Initialization
%  Here we initialize some parameters used for the exercise.

numPatches = 10000;
numFeatures = 196;
patchDim = 14;
visibleSize = patchDim * patchDim;

batchNumPatches = 1000;

outputDir = '.';
lambda = 0.5;  % L1-regularisation parameter (on features)
epsilon = 1e-6; % L1-regularisation epsilon |x| ~ sqrt(x^2 + epsilon)
gamma = 1e-2;   % L2-regularisation parameter (on basis)

%%======================================================================
%% STEP 1: Sample patches

%{
images = load('IMAGES.mat');
images = images.IMAGES;

patches = sampleIMAGES(images, patchDim, numPatches);
display_network(patches(:, 1:64));
%}

patches = load('stdPatches.mat');
patches = patches.patches;

%%======================================================================
%% STEP 2: Sparse coding cost function
%{
numFeatures = 5;
patches = patches(:, 1:5);
numPatches = 5;

weightMatrix = randn(visibleSize, numFeatures) * 0.005;
featureMatrix = randn(numFeatures, numPatches) * 0.005;

theta = [weightMatrix(:); featureMatrix(:); ];

%{
[cost, grad] = sparseCodingCost('both', theta, visibleSize, numFeatures, patches, gamma, lambda, epsilon);

numgrad = computeNumericalGradient( @(x) sparseCodingCost('both', x, visibleSize, numFeatures, patches, gamma, lambda, epsilon), theta );
disp([numgrad grad]); 
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff);
%}

[cost, grad] = sparseCodingWeightCost(weightMatrix, featureMatrix, visibleSize, numFeatures, patches, gamma, lambda, epsilon);

numgrad = computeNumericalGradient( @(x) sparseCodingWeightCost(x, featureMatrix, visibleSize, numFeatures, patches, gamma, lambda, epsilon), weightMatrix(:) );
%disp([numgrad grad]); 
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff);

[cost, grad] = sparseCodingFeatureCost(weightMatrix, featureMatrix, visibleSize, numFeatures, patches, gamma, lambda, epsilon);

numgrad = computeNumericalGradient( @(x) sparseCodingFeatureCost(weightMatrix, x, visibleSize, numFeatures, patches, gamma, lambda, epsilon), featureMatrix(:) );
%disp([numgrad grad]); 
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff);
%}
%%======================================================================
%% STEP 3: Iterative optimization

% minFunc options
options.Method = 'lbfgs';
options.display = 'off';
options.verbose = 0;

% Initialize matrices
weightMatrix = rand(visibleSize, numFeatures) - 0.5;
weightMatrix = bsxfun(@minus, weightMatrix, mean(weightMatrix, 1));
weightMatrix = weightMatrix * diag(1 ./ sqrt(sum(weightMatrix .^2 )) );

indices = randperm(numPatches);
indices = indices(1:batchNumPatches);
batchPatches = patches(:, indices); 
    
featureMatrix = weightMatrix' * batchPatches;
normWM = sum(weightMatrix .^ 2)';
featureMatrix = bsxfun(@rdivide, featureMatrix, normWM);    

theta = [weightMatrix(:); featureMatrix(:); ];
                       
MCLowerBound = [ones(numel(weightMatrix), 1) * -1; ones(numel(featureMatrix), 1) * -Inf ];
MCUpperBound = [ones(numel(weightMatrix), 1) * 1; ones(numel(featureMatrix), 1) * Inf];

%{
for iteration = 1:100
    fprintf('*** Iteration %d\n', iteration);
          
    % Visualize
    
    weightMatrix = reshape(theta(1:visibleSize * numFeatures), visibleSize, numFeatures);
    featureMatrix = reshape(theta(visibleSize * numFeatures + 1:visibleSize * numFeatures + numFeatures * batchNumPatches), numFeatures, batchNumPatches);

    error = weightMatrix * featureMatrix - batchPatches;
    error = sum(error(:) .^ 2);
    
    fResidue = error;
    fSparsity = lambda * sum(sqrt(featureMatrix(:) .^ 2 + epsilon));
    fWeight = 1 * sum(weightMatrix(:) .^ 2);
    
    fprintf('%-12s%-12s%-12s%-12s\n','fObj','fResidue','fSparsity','fWeight');
    fprintf('%6.4f  %6.4f  %6.4f  %6.4f\n', fResidue+fSparsity, fResidue, fSparsity, fWeight)
            
    %display_network(patches(:, 1:36));
    %figure(1);
    %display_network(weightMatrix(:, 1:64));
    
    indices = randperm(numPatches);
    indices = indices(1:batchNumPatches);
    batchPatches = patches(:, indices);    
    
    featureMatrix = weightMatrix' * batchPatches;
    normWM = sum(weightMatrix .^ 2)';
    featureMatrix = bsxfun(@rdivide, featureMatrix, normWM); 
    
    fSparsity = lambda * sum(sqrt(featureMatrix(:) .^ 2 + epsilon));
    fprintf('%6.4f\n', fSparsity);
    
    theta = [weightMatrix(:); featureMatrix(:); ];

    % Optimize for feature matrix
    
    options.maxIter = 100;
    [optTheta, cost] = minFunc( @(x) sparseCodingCost('feature', x, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
                                    theta, options);

    theta = optTheta;    
    disp(cost);
    
    % Optimize for weight matrix
       
    weightMatrix = reshape(theta(1:visibleSize * numFeatures), visibleSize, numFeatures);
    featureMatrix = reshape(theta(visibleSize * numFeatures + 1:visibleSize * numFeatures + numFeatures * batchNumPatches), numFeatures, batchNumPatches);

    weightMatrix = rand(visibleSize, numFeatures) - 0.5;
    weightMatrix = bsxfun(@minus, weightMatrix, mean(weightMatrix, 1));
    weightMatrix = weightMatrix * diag(1 ./ sqrt(sum(weightMatrix .^2 )) );    

    theta = [weightMatrix(:); featureMatrix(:); ];
       
    options.maxIter = 20;
    
    %{
    [optTheta, cost] = minConf_TMP( @(x) sparseCodingCost('weight', x, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
                                    theta, MCLowerBound, MCUpperBound, options);                           
    %}    
    [optTheta, cost] = minConf_SPG( @(x) sparseCodingCost('weight', x, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
                                    theta, @(x) sparseCodingProjection('restrict', x, visibleSize, numFeatures, batchNumPatches), options);                                      
    %{
    [optTheta, cost] = minFunc( @(x) sparseCodingCost('weight', x, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
                                    theta, options);                           
    %}
                           
    theta = optTheta;
    disp(cost);    
    
   
end
%}

warning('off', 'MATLAB:nearlySingularMatrix')

fprintf('%6s%12s%12s%12s%12s\n','Iter', 'fObj','fResidue','fSparsity','fWeight');

startTime = tic();

for iteration = 1:500
%     fprintf('*** Iteration %d\n', iteration);
          
    % Visualize        
    
    error = weightMatrix * featureMatrix - batchPatches;
    error = sum(error(:) .^ 2);
    
    fResidue = error;
    fSparsity = lambda * sum(sqrt(featureMatrix(:) .^ 2 + epsilon));
    fWeight = gamma * sum(weightMatrix(:) .^ 2);
    
    fprintf('  %4d  %10.4f  %10.4f  %10.4f  %10.4f\n', iteration, fResidue+fSparsity+fWeight, fResidue, fSparsity, fWeight)
            
%     display_network(patches(:, 1:36));
%     figure(1);
%     display_network(weightMatrix(:, 1:64));
    
    indices = randperm(numPatches);
    indices = indices(1:batchNumPatches);
    batchPatches = patches(:, indices);                    
    
    featureMatrix = weightMatrix' * batchPatches;
    normWM = sum(weightMatrix .^ 2)';
    featureMatrix = bsxfun(@rdivide, featureMatrix, normWM); 
    
    fSparsity = lambda * sum(sqrt(featureMatrix(:) .^ 2 + epsilon));
    
    theta = [weightMatrix(:); featureMatrix(:); ];

    % Optimize for feature matrix
    
%     options.maxIter = 20;
%     [featureMatrix, cost] = minFunc( @(x) sparseCodingFeatureCost(weightMatrix, x, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
%                                           featureMatrix(:), options);
%     featureMatrix = reshape(featureMatrix, numFeatures, batchNumPatches);                                      
    
    
    [featureMatrix] = admmLasso(weightMatrix, featureMatrix, featureMatrix, 0, batchPatches, numFeatures, visibleSize, lambda, 30, 30);
    
    % Optimize for weight matrix
       
%     weightMatrix = rand(visibleSize, numFeatures) - 0.5;
%     weightMatrix = bsxfun(@minus, weightMatrix, mean(weightMatrix, 1));
%     weightMatrix = weightMatrix * diag(1 ./ sqrt(sum(weightMatrix .^2 )) );    
%            
%     options.maxIter = 20;
% 
%     %{
%     [weightMatrix, cost] = minConf_SPG( @(x) sparseCodingWeightCost(x, featureMatrix, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
%                                              weightMatrix(:), @(x) sparseCodingWeightProjection('restrict', weightMatrix(:), visibleSize, numFeatures, batchNumPatches), options);
%     %}
%     
%     
%     [weightMatrix, cost] = minFunc( @(x) sparseCodingWeightCost(x, featureMatrix, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon), ...
%                                     weightMatrix(:), options);
%     
% 
%     weightMatrix = reshape(weightMatrix, visibleSize, numFeatures);                                         
%     
    weightMatrix = batchPatches * featureMatrix' / batchNumPatches * inv(featureMatrix*featureMatrix' / batchNumPatches + gamma/2*eye(numFeatures));
    
%     disp([sparseCodingWeightCost(weightMatrix(:), featureMatrix, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon) sparseCodingWeightCost(myWeightMatrix(:), featureMatrix, visibleSize, numFeatures, batchPatches, gamma, lambda, epsilon)]);

%    fprintf('cost = %10.4f\n', cost);

    if mod(iteration, 1) == 0
        duration = toc(startTime);
        % save(sprintf('%s/sparseCodingTheta-i%d-%g-%g-%g.mat', outputDir, iteration, lambda, epsilon, gamma), 'weightMatrix', 'featureMatrix', 'duration');

        display_network(weightMatrix(:,:));
        % print('-dpng',sprintf('%s/sparseCodingTheta-i%d-%g-%g-%g.png', outputDir, iteration, lambda, epsilon, gamma) );        
    end
   
end
