function [projectedTheta] = sparseCodingProjection(which, theta, visibleSize, numFeatures, numExamples)
%sparseCodingProjection - returns theta with the norm of all basis vectors
%                         modified as given by which
% Parameters
%   which - 'fix' - set norms of all basis vectors to targetNorm
%           'restrict' - limit norms of all basis vectors to at most targetNorm
%                        (i.e. set norms of all basis vectors which have
%                         norm > targetNorm to targetNorm, but leave other 
%                         basis vectors unchanged)
%   
    targetNorm = 1;

    weightMatrix = reshape(theta(1:visibleSize * numFeatures), visibleSize, numFeatures);
    featureMatrix = reshape(theta(visibleSize * numFeatures + 1:visibleSize * numFeatures + numFeatures * numExamples), numFeatures, numExamples);
    
    temp = sqrt(sum(weightMatrix .^ 2, 1));
    if isequal(which, 'restrict') 
        % Leave basis vectors with norm less than targetNorm unchanged
        temp(temp < targetNorm) = targetNorm;
    end
    temp = targetNorm ./ temp;    
    
    weightMatrix = weightMatrix * diag(temp);
    
    projectedTheta = [weightMatrix(:); featureMatrix(:)];    
end