function [cost, grad] = sparseCodingWeightCost(weightMatrix, featureMatrix, visibleSize, numFeatures,  patches, gamma, lambda, epsilon)
%sparseCodingWeightCost 

    numExamples = size(patches, 2);

    weightMatrix = reshape(weightMatrix, visibleSize, numFeatures);
    featureMatrix = reshape(featureMatrix, numFeatures, numExamples);

    % Reconstruction error
    cost = weightMatrix * featureMatrix - patches;
    cost = sum(cost(:) .^ 2) / numExamples; 
    
    % Be careful! This is a differentiable form of L1-penalty 
    % (__NOT__ an L2 penalty), hence the sqrt
    sparsityPenaltyMatrix = sqrt(featureMatrix .^ 2 + epsilon);
    sparsityPenalty = sum(sparsityPenaltyMatrix(:));
    
    % Basis weight penalty
    cost = cost + (gamma / 2) * sum(weightMatrix(:) .^ 2);
    
    % Feature sparsity penalty
    cost = cost + lambda * sparsityPenalty;    
    
    delta = weightMatrix * featureMatrix - patches;
    weightGrad = 2 * delta * featureMatrix' / numExamples; 
    weightGrad = weightGrad + gamma * weightMatrix;
    featureGrad = 2 * weightMatrix' * delta;    
    featureGrad = featureGrad + lambda * featureMatrix ./ sparsityPenaltyMatrix;
    
    grad = weightGrad(:);     
end