function patches = sampleIMAGES(IMAGES, patchSize, numPatches)
global AEOptions;

% sampleIMAGES
% Returns 10000 patches for training

% Initialize patches with zeros.  Your code will fill in this matrix--one
% column per patch, 10000 columns. 
patches = zeros(patchSize*patchSize, numPatches);

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Fill in the variable called "patches" using data 
%  from IMAGES.  
%  
%  IMAGES is a 3D array containing 10 images
%  For instance, IMAGES(:,:,6) is a 512x512 array containing the 6th image,
%  and you can type "imagesc(IMAGES(:,:,6)), colormap gray;" to visualize
%  it. (The contrast on these images look a bit off because they have
%  been preprocessed using using "whitening."  See the lecture notes for
%  more details.) As a second example, IMAGES(21:30,21:30,1) is an image
%  patch corresponding to the pixels in the block (21,21) to (30,30) of
%  Image 1


% ------------ Sample Solution ------------
p = 0;
for im = 1:size(IMAGES, 3)
    
    % Sample Patches
    numsamples = numPatches / size(IMAGES, 3);
    for s = 1:numsamples
        acceptable = false;
        
        % DEBUG (for photos): 
        % Reject low-variance images (to avoid entirely white patches); 
        % Reject high mean images (to avoid patches of sky);
        p = p + 1;        
        while ~acceptable            
            y = randi(size(IMAGES,1)-patchSize+1);
            x = randi(size(IMAGES,2)-patchSize+1);
            sample = IMAGES(y:y+patchSize-1, x:x+patchSize-1,im);            
            patches(:, p) = sample(:);
            if ((var(sample(:)) < 1e-5) || (mean(sample(:)) > 256))
                acceptable = false;
            else                
                acceptable = true;
            end
        end
        
    end
   
end
% ---------- End Sample Solution ---------- 

end


%% ---------------------------------------------------------------
function patches = normalizeData(patches)

% Squash data to [0.1, 0.9] since we use sigmoid as the activation
% function in the output layer

% Remove DC (mean of images). 
patches = bsxfun(@minus, patches, mean(patches));

% Truncate to +/-3 standard deviations and scale to -1 to 1
pstd = 3 * std(patches(:));
patches = max(min(patches, pstd), -pstd) / pstd;

% Rescale from [-1,1] to [0.1,0.9]
patches = (patches + 1) * 0.4 + 0.1;

end
