function [ell, deriv, pred] = softmaxClassifierCost(incoming, weights, outputMatrix)
%softmaxClassifierCost 
%   incoming - vector of incoming values 
%              (incoming(k,1) = theta_k x^(i) for all k (i.e. all classes))
%   weights - matrix containing thetas, such that weights(r, c) = theta_r, c
%   outputMatrix - otuputMatrix(r,c) = 1 if y(c) = r; 0 otherwise
%
%   ell   - mean log-likelihood (not negated, stackedAutoencoderCost negates it)
%   deriv - matrix of mean derivatives of log-likelihood wrt thetas
%   pred  - matrix of predictions

M = size(outputMatrix, 2);

zClassifier = incoming;
zClassifier = bsxfun(@minus, zClassifier, max(zClassifier,[],1)); % Rescale
expClassifier = exp(zClassifier);

ell = 0;
zTerm = (zClassifier .* outputMatrix);
zTerm = sum(zTerm(:));
ell = ell + ( zTerm - sum(log(sum(expClassifier, 1))) );
ell = ell / M;

pred = bsxfun(@rdivide, expClassifier, sum(expClassifier, 1));
deriv = outputMatrix - pred;

deriv = deriv / M;

end

