%% CS294A/CS294W Programming Assignment Starter Code
%% Training for SOFTMAX

%%======================================================================
%% STEP 0: Here we provide the relevant parameters values that will
%  allow your sparse autoencoder to get good filters; you do not need to 
%  change the parameters below.

dataType = 'features3'; % pixels or features
outputSize = 10;
softmaxLambda = 1e-3;

if isequal(dataType, 'pixels')
    inputSize = 28 * 28;
elseif isequal(dataType, 'features1')
    inputSize = 500;
elseif isequal(dataType, 'features2')
    inputSize = 500;
elseif isequal(dataType, 'features3')
    inputSize = 1000;
end

%%======================================================================
%% STEP 1: Implement sampleIMAGES
%
%  After implementing sampleIMAGES, the display_network command should
%  display a random sample of 200 patches from the dataset

if isequal(dataType, 'pixels') 

    images = loadMNISTImages('mnist/train-images-idx3-ubyte');
    labels = loadMNISTLabels('mnist/train-labels-idx1-ubyte');
    labels(labels==0) = 10; % Remap 0 to 10

    patches = reshape(images, 28 * 28, size(images, 3));
    patches = patches / 255;
    display_network(patches(:,randi(size(patches,2),200,1)),8);
    
    inputData = patches;
    outputData = labels;        
elseif isequal(dataType(1:8), 'features')
    
    % NOTE: Run step 6 of train.m first!
    
    inputData = features;
    outputData = labels;        
end

inputData = inputData(:, 1:10000);
outputData = outputData(1:10000);

theta = 0.005 * randn(outputSize * inputSize + outputSize, 1);

%%======================================================================
%% STEP 2: Implement sparseAutoencoderCost
%
[cost, grad, pred] = softmaxCost(theta, outputSize, inputSize, softmaxLambda, inputData, outputData);
                                     
%%======================================================================
%% STEP 3: Gradient Checking
%
% Hint: If you are debugging your code, performing gradient checking on smaller models 
% and smaller training sets (e.g., using only 10 training examples and 1-2 hidden 
% units) may speed things up.

% First, lets make sure your numerical gradient computation is correct for a
% simple function.  After you have implemented computeNumericalGradient.m,
% run the following: 
checkNumericalGradient();

% Now we can use it to check your cost function and derivative calculations
% for the sparse autoencoder.  
numgrad = computeNumericalGradient( @(x) softmaxCost(x, outputSize, ...
                                    inputSize, softmaxLambda, inputData, outputData), theta);

% Use this to visually compare the gradients side by side
disp([numgrad grad]); 

% Compare numerically computed gradients with the ones obtained from backpropagation
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff); % Should be small. In our implementation, these values are
            % usually less than 1e-9.

            % When you got this working, Congratulations!!! 

%%======================================================================
%% STEP 4: L-BFGS
%

%  Randomly initialize the parameters
theta = 0.005 * randn(outputSize * inputSize + outputSize, 1);

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 100;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
options.outputFcn = '';

[softmaxOptTheta, cost] = minFunc( @(p) softmaxCost(p, ...
                                   outputSize, inputSize, softmaxLambda, ...
                                   inputData, outputData), ...                                   
                              theta, options);
                          
save(['softmaxOptTheta-',dataType,'.mat'],'softmaxOptTheta','cost');

%%======================================================================
%% STEP 5: Cross-validation

softmaxOptTheta = load(['softmaxOptTheta-',dataType,'.mat']);
softmaxOptTheta = softmaxOptTheta.softmaxOptTheta;
%{
% Cross-validate against test set
images = loadMNISTImages('mnist/t10k-images-idx3-ubyte');
labels = loadMNISTLabels('mnist/t10k-labels-idx1-ubyte');
labels(labels==0) = 10; % Remap 0 to 10

patches = reshape(images, 28 * 28, size(images, 3));
patches = patches / 255;
% display_network(patches(:,randi(size(patches,2),200,1)),8);

if isequal(dataType, 'pixels') 
    inputData = patches;
    outputData = labels;
elseif isequal(dataType, 'features') 
    AEOptTheta = load('AEOptThetaPart.mat');
    AEOptTheta = AEOptTheta.AEOptTheta;

    % NOTE:
    % Need to initialise these values by running sparseAutoencoderCost.m first
    [cost, grad, features] = sparseAutoencoderCost(AEOptTheta, ...
                                visibleSize, hiddenSize, ...
                                lambda, sparsityParam, beta, patches);
    inputData = features;
    outputData = labels;
    
end
%}

inputData = features;
outputData = labels;

%%
                          
[cost, grad, pred] = softmaxCost(softmaxOptTheta, outputSize, inputSize, softmaxLambda, inputData, outputData);
outputMatrix = full(sparse(outputData,1:size(inputData,2),1));
error = (1/2) * norm(pred - outputMatrix) ^ 2;
error = error / size(inputData, 2);
[prob, pred] = max(pred);
pred = pred';

acc = outputData - pred;
acc = acc(acc==0);
acc = size(acc, 1) / size(outputData,1);
fprintf('Accuracy: %0.6f\n', acc);
fprintf('Error: %0.6f\n', error);