%% CS294A/CS294W Programming Assignment Starter Code
%% ======================================================================
%% STACKED AE code

%%======================================================================
%% STEP 0: Here we provide the relevant parameters values that will
%  allow your sparse autoencoder to get good filters; you do not need to 
%  change the parameters below.

global startTime;
global outputDir;
global AEOptions;

outputDir = './scratch/';

patchSize = 28;        % patch dimension
numPatches = 60000;    % number of patches

visibleSize = patchSize * patchSize;   % number of input units 
outputSize = visibleSize;              % number of output units
hiddenSize = 196;                      % number of hidden units 

sparsityParam = 0.1;   % desired average activation of the hidden units.
                       % (This was denoted by the Greek alphabet rho, which looks like a lower-case "p",
                       %  in the lecture notes). 
lambda = 3e-3;         % weight decay parameter       
beta = 3;              % weight of sparsity penalty term       

sizeVector = [visibleSize; hiddenSize; outputSize];

%%======================================================================
%% STEP 1: Implement sampleIMAGES
%
%  After implementing sampleIMAGES, the display_network command should
%  display a random sample of 200 patches from the dataset

TRAIN_SET = false;
if TRAIN_SET   
    disp('Using TRAINING set');
    images = loadMNISTImages('mnist/train-images-idx3-ubyte');
    labels = loadMNISTLabels('mnist/train-labels-idx1-ubyte');
else 
    disp('Using TEST set');
    images = loadMNISTImages('mnist/t10k-images-idx3-ubyte');
    labels = loadMNISTLabels('mnist/t10k-labels-idx1-ubyte');
end

labels(labels==0) = 10; % Remap 0 to 10

patches = reshape(images, 28 * 28, size(images, 3));
patches = patches / 255;

% Take only 0-4
%{
part = find((labels <= 4) | (labels == 10));
disp(part);
patches = patches(:, part);
labels = labels(part);
%}
% End take only 0-4

display_network(patches(:,randi(size(patches,2),200,1)),8);

% Use less than all the patches for debugging / speed
patches = patches(:, 1:numPatches);
labels = labels(1:numPatches);

%  Obtain random parameters theta
theta = stackedInitializeParameters(sizeVector);

% Check rolling and unrolling functions
stackedRollCheck();

%%======================================================================
%% STEP 2: Implement stackedAutoencoderCost
%

[cost, grad] = stackedAutoencoderCost(theta, sizeVector, 'sparseAE', lambda, ...
                                      sparsityParam, beta, patches, labels);                                                                   

%%======================================================================
%% STEP 3: Gradient Checking
%

checkNumericalGradient();

numgrad = computeNumericalGradient( @(x) stackedAutoencoderCost(x, sizeVector, 'sparseAE', ...
                                                  lambda, sparsityParam, beta, ...
                                                  patches, labels), theta);

% Use this to visually compare the gradients side by side
disp([numgrad grad]); 

% Compare numerically computed gradients with the ones obtained from backpropagation
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff); % Should be small. In our implementation, these values are
            % usually less than 1e-9.
            
%%======================================================================
%% STEP 4: L-BFGS, layer-wise

% which hidden layer are we currently training 
% for the standard autoencoder, currentHidden = 2
% for the 2-layer autoencoder, currentHidden = 3
% for the 3-layer autoencoder, currentHidden = 4, etc
currentHidden = 5; 
hiddenSizeVector = [28*28; 500; 500; 1000; 10];

if currentHidden == 2
    features = patches;
else
    
    % partialSizeVector for getting the features for training this new hidden layer
    partialSizeVector = [hiddenSizeVector(1:currentHidden-1); hiddenSizeVector(currentHidden-2)];

    partialTheta = stackedInitializeParameters(partialSizeVector);
    [~, partialW, partialb] = stackedRollOutOfTheta(partialSizeVector, partialTheta);

    % Load weights for earlier layers
    for i = 2:currentHidden-1            
        tempSizeVector = [hiddenSizeVector(i-1); hiddenSizeVector(i); hiddenSizeVector(i-1)];
        temp = load(sprintf('stacked-optTheta-layer%02d.mat', i));
        temp = temp.AEOptTheta;        
        [~, tempW, tempb] = stackedRollOutOfTheta(tempSizeVector, temp);            
        partialW{i-1} = tempW{1};
        partialb{i-1} = tempb{1};
    end    
    partialW{i} = tempW{2};
    partialb{i} = tempb{2};        
        
    partialTheta = stackedRollIntoTheta(partialSizeVector, partialW, partialb);
        
    % Get features from raw data, using the partial autoencoder
    [~, ~, activation] = stackedAutoencoderCost(partialTheta, partialSizeVector, 'sparseAE', ...
                                                lambda, sparsityParam, beta, ...
                                                patches, labels);
    features = activation{currentHidden-1};    
end

%%

% Okay, now prepare THIS hidden layer for training
thisSizeVector = [hiddenSizeVector(currentHidden-1); hiddenSizeVector(currentHidden); hiddenSizeVector(currentHidden-1)];

% Set LOAD to true to load saved parameters
LOAD = true;
if LOAD
    % Load saved parameters
    temp = load(sprintf('stacked-optTheta-layer%02d.mat', currentHidden));    
    temp = temp.AEOptTheta;
    thisTheta = temp;
else
    % Reinitialise
    thisTheta = stackedInitializeParameters(thisSizeVector);
end

assert(TRAIN_SET == true, 'Are you sure you want to train on the test set?');

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 400;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
options.outputFcn = @AELbfgsProgress;

startTime = tic;
[AEOptTheta, cost] = minFunc( @(p) stackedAutoencoderCost(p, ...
                                   thisSizeVector, 'sparseAE', ...
                                   lambda, sparsityParam, ...
                                   beta, features, labels), ...
                              thisTheta, options);

save(sprintf('stacked-optTheta-layer%02d.mat', currentHidden),'AEOptTheta','cost','currentHidden','sizeVector');
                          
%%======================================================================
%% STEP 5: Visualization 

AEOptTheta = load(sprintf('stacked-optTheta-layer%02d.mat', 2));
AEOptTheta = AEOptTheta.AEOptTheta;

W1 = reshape(AEOptTheta(1:hiddenSizeVector(2)*hiddenSizeVector(1)), hiddenSizeVector(2), hiddenSizeVector(1));
display_network(W1', 12); 

%%

% NOTE: Run step 4, then this step (to set currentHidden), 
%       then 6a, to initialise FTTheta appropriately

currentHidden = 3;
%%

[~, tW, tb] = stackedRollOutOfTheta(FTSizeVector, FTTheta);
FTTheta = stackedRollIntoTheta(FTSizeVector, tW, tb);    


%% STEP 5a: Gradient checking for visualisation

FTSizeVector = [10;20;30];
FTTheta = stackedInitializeParameters(FTSizeVector);
randomInput = rand(FTSizeVector(1), 1);

[cost, grad] = stackedAutoencoderCost(FTTheta, FTSizeVector, @visualisationClassifierCost, ...
                                      lambda, sparsityParam, beta, randomInput, labels);                                                                   

numgrad = computeNumericalGradient( @(x) stackedAutoencoderCost(FTTheta, FTSizeVector, ...
                                            @visualisationClassifierCost, ...
                                            lambda, sparsityParam, beta, ...
                                            x, labels), randomInput);
    
% Use this to visually compare the gradients side by side
disp([numgrad grad]); 

% Compare numerically computed gradients with the ones obtained from backpropagation
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff); % Should be small. In our implementation, these values are
            % usually less than 1e-9.
            
%% STEP 5b: Get visualisation      

global Z_VCWHICH;

Z_IN = zeros(28*28, 100);

for Z_VCWHICH = 1:30

disp(Z_VCWHICH);    
    
randomInput = patches(:, randi([1, 1000]));

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 100;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
options.outputFcn = '';
options.TolFun = 1e-12;
options.TolX = 1e-12;
options.Display = 'off';

[i, c] = stackedAutoencoderCost(FTTheta, ...
                                   FTSizeVector, @visualisationClassifierCost, ...
                                   lambda, sparsityParam, ...
                                   beta, randomInput, labels);
                          
if c > -1e-4
    % If we don't get off to a good start, then we might as well give up
    continue;
end;

[optInput, cost] = minFunc( @(p) stackedAutoencoderCost(FTTheta, ...
                                   FTSizeVector, @visualisationClassifierCost, ...
                                   lambda, sparsityParam, ...
                                   beta, p, labels), ...
                              randomInput, options);

Z_IN(:,Z_VCWHICH) = optInput;                          
                          
end

display_network(Z_IN(:,1:Z_VCWHICH));  

%%======================================================================
%% STEP 6: L-BFGS, fine-tuning

%% STEP 6a: Initialisation

% Fine-tune the entire network together

FTSizeVector = hiddenSizeVector(1:currentHidden);

% Set LOAD to true to load saved parameters
LOAD = true;
if LOAD
    % Load saved parameters
    FTTheta = load('stacked-optTheta-fineTuned.mat');
    FTTheta = FTTheta.FTOptTheta;    
else        
    % Reinitialise
    % NOTE: load partialTheta, partialSizeVector using step 4 first
    % NOTE: Have to roll and unroll because loading using step 4 means 
    %       that we will load an extra layer at the end that we don't want
    error('Are you sure you want to reinitialise theta?');
    FTTheta = partialTheta;
    [~, tW, tb] = stackedRollOutOfTheta(FTSizeVector, FTTheta);
    FTTheta = stackedRollIntoTheta(FTSizeVector, tW, tb);
end

%% STEP 6b: L-BFGS

assert(TRAIN_SET == true, 'Are you sure you want to train on the test set?');

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 100;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
options.outputFcn = @AELbfgsProgress;

startTime = tic;
[FTOptTheta, cost] = minFunc( @(p) stackedAutoencoderCost(p, ...
                                   FTSizeVector, @softmaxClassifierCost, ...
                                   lambda, sparsityParam, ...
                                   beta, patches, labels), ...
                              FTTheta, options);

save('stacked-optTheta-fineTuned.mat', 'FTOptTheta','cost','currentHidden','FTSizeVector');
        
%% STEP 6c: Cross-validation

[~, ~, ~, pred] = stackedAutoencoderCost(FTTheta, FTSizeVector, @softmaxClassifierCost, ...
                                         lambda, sparsityParam, beta, ...
                                         patches, labels);
                                     
[prob, pred] = max(pred);
acc = (pred' == labels);
acc = sum(acc) / size(acc, 1);
fprintf('Accuracy: %0.6f\n', acc);