%% CS294A/CS294W Programming Assignment Starter Code

%  Instructions
%  ------------
% 
%  This file contains code that helps you get started on the
%  programming assignment. You will need to complete the code in sampleIMAGES.m,
%  sparseAutoencoderCost.m and computeNumericalGradient.m. 
%  For the purpose of completing the assignment, you do not need to
%  change the code in this file. 
%
%%======================================================================
%% STEP 0: Here we provide the relevant parameters values that will
%  allow your sparse autoencoder to get good filters; you do not need to 
%  change the parameters below.

global visibleSize hiddenSize;
global startTime;
global outputDir;
global AEOptions;

AEOptions.linearAutoencoder = false;

outputDir = './scratch/';
patchSize = 28;     % patch dimension
numPatches = 10000; % number of patches
visibleSize = patchSize * patchSize;   % number of input units 
outputSize = 10;    % number of output units
hiddenSize = 196;   % number of hidden units 
sparsityParam = 0.1;   % desired average activation of the hidden units.
                       % (This was denoted by the Greek alphabet rho, which looks like a lower-case "p",
                       %  in the lecture notes). 
lambda = 3e-3;      % weight decay parameter       
beta = 3;           % weight of sparsity penalty term       

%%======================================================================
%% STEP 1: Implement sampleIMAGES
%
%  After implementing sampleIMAGES, the display_network command should
%  display a random sample of 200 patches from the dataset

images = loadMNISTImages('mnist/train-images-idx3-ubyte');
labels = loadMNISTLabels('mnist/train-labels-idx1-ubyte');
labels(labels==0) = 10; % Remap 0 to 10

patches = reshape(images, 28 * 28, size(images, 3));
patches = patches / 255;

% Take only 0-4
%{
part = find((labels <= 4) | (labels == 10));
disp(part);
patches = patches(:, part);
labels = labels(part);
%}
% End take only 0-4

display_network(patches(:,randi(size(patches,2),200,1)),8);

% Use less than all the patches for debugging / speed
patches = patches(:, 1:numPatches);
labels = labels(1:numPatches);

%  Obtain random parameters theta
theta = combinedInitializeParameters(visibleSize, hiddenSize, visibleSize);

%%======================================================================
%% STEP 2: Implement sparseAutoencoderCost
%
%  You can implement all of the components (squared error cost, weight decay term,
%  sparsity penalty) in the cost function at once, but it may be easier to do 
%  it step-by-step and run gradient checking (see STEP 3) after each step.  We 
%  suggest implementing the sparseAutoencoderCost function using the following steps:
%
%  (a) Implement forward propagation in your neural network, and implement the 
%      squared error term of the cost function.  Implement backpropagation to 
%      compute the derivatives.   Then (using lambda=beta=0), run Gradient Checking 
%      to verify that the calculations corresponding to the squared error cost 
%      term are correct.
%
%  (b) Add in the weight decay term (in both the cost function and the derivative
%      calculations), then re-run Gradient Checking to verify correctness. 
%
%  (c) Add in the sparsity penalty term, then re-run Gradient Checking to 
%      verify correctness.
%
%  Feel free to change the training settings when debugging your
%  code.  (For example, reducing the training set size or 
%  number of hidden units may make your code run faster; and setting beta 
%  and/or lambda to zero may be helpful for debugging.)  However, in your 
%  final submission of the visualized weights, please use parameters we 
%  gave in Step 0 above.

[cost, grad] = sparseAutoencoderCost(theta, visibleSize, hiddenSize, lambda, ...
                                     sparsityParam, beta, patches);

%%======================================================================
%% STEP 3: Gradient Checking
%
% Hint: If you are debugging your code, performing gradient checking on smaller models 
% and smaller training sets (e.g., using only 10 training examples and 1-2 hidden 
% units) may speed things up.

% First, lets make sure your numerical gradient computation is correct for a
% simple function.  After you have implemented computeNumericalGradient.m,
% run the following: 
checkNumericalGradient();

% Now we can use it to check your cost function and derivative calculations
% for the sparse autoencoder.  
numgrad = computeNumericalGradient( @(x) sparseAutoencoderCost(x, visibleSize, ...
                                                  hiddenSize, lambda, ...
                                                  sparsityParam, beta, ...
                                                  patches), theta);

% Use this to visually compare the gradients side by side
disp([numgrad grad]); 

% Compare numerically computed gradients with the ones obtained from backpropagation
diff = norm(numgrad-grad)/norm(numgrad+grad);
disp(diff); % Should be small. In our implementation, these values are
            % usually less than 1e-9.

            % When you got this working, Congratulations!!! 

%%======================================================================
%% STEP 4: After verifying that your implementation of
%  sparseAutoencoderCost is correct, You can start training your sparse
%  autoencoder with minFunc (L-BFGS).

%  Randomly initialize the parameters
theta = combinedInitializeParameters(visibleSize, hiddenSize, visibleSize);

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 400;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
%options.outputFcn = @AELbfgsProgress;

startTime = tic;
[AEOptTheta, cost] = minFunc( @(p) sparseAutoencoderCost(p, ...
                                   visibleSize, hiddenSize, ...
                                   lambda, sparsityParam, ...
                                   beta, patches), ...
                              theta, options);

%save('AEOptThetaPart.mat','AEOptTheta','cost');
                          
%%======================================================================
%% STEP 5: Visualization 

AEOptTheta = load('AEOptThetaPart.mat');
AEOptTheta = AEOptTheta.AEOptTheta;

W1 = reshape(AEOptTheta(1:hiddenSize*visibleSize), hiddenSize, visibleSize);
display_network(W1', 12); 

print -djpeg weights.jpg   % save the visualization to a file 

%% =====================================================================
%% ==================           Softmax                =================
%% =====================================================================

%%======================================================================
%% STEP 6: Dump features for training softmax

[cost, grad, features] = sparseAutoencoderCost(AEOptTheta, ...
                            visibleSize, hiddenSize, ...
                            lambda, sparsityParam, ...
                            beta, patches);
                        
softmaxData.features = features;
softmaxData.labels = labels;

%%======================================================================
%% STEP 7: Train softmax classification on features using softmaxTrain.m

% Use softmaxTrain.m

%% =====================================================================
%% ==================           Combined               =================
%% =====================================================================

%%======================================================================
%% STEP 8: Initialise combined autoencoder

AEOptTheta = load('AEOptTheta.mat');
AEOptTheta = AEOptTheta.AEOptTheta;

% AE
W1 = reshape(AEOptTheta(1:hiddenSize*visibleSize), hiddenSize, visibleSize);
b1 = AEOptTheta(2*hiddenSize*visibleSize+1:2*hiddenSize*visibleSize+hiddenSize);

softmaxOptTheta = load('softmaxOptTheta-features.mat');
softmaxOptTheta = softmaxOptTheta.softmaxOptTheta;

% softmax
W2 = reshape(softmaxOptTheta(1:outputSize*hiddenSize), outputSize, hiddenSize);
temp = outputSize * hiddenSize;
b2 = softmaxOptTheta(temp+1:temp+outputSize);

theta = [W1(:); W2(:); b1(:); b2(:)];

%%======================================================================
%% Step 9: Train combined autoencoder with L-BFGS

% WARNING: Run step 1 again to initialise images and labels!

% WARNING: Run step 8 to initialise theta first!

%  Use minFunc to minimize the function
addpath minFunc/
options.Method = 'lbfgs'; % Here, we use L-BFGS to optimize our cost
                          % function. Generally, for minFunc to work, you
                          % need a function pointer with two outputs: the
                          % function value and the gradient. In our problem,
                          % sparseAutoencoderCost.m satisfies this.
options.maxIter = 100;	  % Maximum number of iterations of L-BFGS to run 
options.display = 'on';
options.outputFcn = '';

startTime = tic;
[combinedOptTheta, cost] = minFunc( @(p) combinedAutoencoderCost(p, ...
                                   outputSize, visibleSize, hiddenSize, ...
                                   lambda, sparsityParam, ...
                                   beta, patches, labels), ...
                              theta, options);

save('combinedOptTheta.mat','combinedOptTheta','cost');

%%======================================================================
%% Step 10: Cross-validation

combinedOptTheta = load('combinedOptTheta.mat');
combinedOptTheta = combinedOptTheta.combinedOptTheta;

% Cross-validate against test set
images = loadMNISTImages('mnist/t10k-images-idx3-ubyte');
labels = loadMNISTLabels('mnist/t10k-labels-idx1-ubyte');
labels(labels==0) = 10; % Remap 0 to 10

patches = reshape(images, 28 * 28, size(images, 3));
patches = patches / 255;
% display_network(patches(:,randi(size(patches,2),200,1)),8);
                        
[cost, grad, features, pred] = combinedAutoencoderCost(combinedOptTheta, ...
                        outputSize, visibleSize, hiddenSize, ...
                        lambda, sparsityParam, beta, ...
                        patches, labels);    
outputMatrix = full(sparse(outputData,1:size(inputData,2),1));
error = (1/2) * norm(pred - outputMatrix) ^ 2;
error = error / size(inputData, 2);
[prob, pred] = max(pred);
pred = pred';

acc = outputData - pred;
acc = acc(acc==0);
acc = size(acc, 1) / size(outputData,1);
fprintf('Accuracy: %0.6f\n', acc);
fprintf('Error: %0.6f\n', error);